//  ======================================================================== //
//                                                                           //
//    File      : SongContainer.h                                            //
//    Purpose   :                                                            //
//    Time-stamp: <Sun Feb/02/2003 09:25 MET Coder@ReallySoft.de>            //
//                                                                           //
//    (C) April 2002 by Ralf Westram                                         //
//                                                                           //
//    Permission to use, copy, modify, distribute and sell this software     //
//    and its documentation for any purpose is hereby granted without fee,   //
//    provided that the above copyright notice appear in all copies and      //
//    that both that copyright notice and this permission notice appear      //
//    in supporting documentation.                                           //
//                                                                           //
//    Ralf Westram makes no representations about the suitability of this    //
//    software for any purpose.  It is provided "as is" without express or   //
//    implied warranty.                                                      //
//                                                                           //
//  ======================================================================== //

#ifndef SONGCONTAINER_H
#define SONGCONTAINER_H

#ifndef SONG_H
#include "Song.h"
#endif


namespace playlist {

    void initPredicates(const std::string& extension_list);

    template <class Container>
    // --------------------------------------------------------------------------------
    //      inline void COPY_ALL(const Container& to_append, Container& append_to)
    // --------------------------------------------------------------------------------
    inline void COPY_ALL(const Container& to_append, Container& append_to) {
        if (!to_append.empty()) {
            if (append_to.empty()) {
                append_to = to_append;
            }
            else {
                copy(to_append.begin(), to_append.end(), std::back_insert_iterator<Container>(append_to));
            }
        }
    }

    // -----------------------------
    //      class SongContainer
    // -----------------------------
    class                               SongContainer;
    typedef rs::SmartPtr<SongContainer> SongContainerPtr;
    typedef std::list<SongContainerPtr> SongContainerList;

    class SongContainer {
    private:
        SongContainerList    sons;
        SongList             songs;
        bool                 force_sorted;
        bool                 force_shuffled;
        const SongContainer *father; // if existing
        std::string               name;

        static std::string all_name;
        static std::string rest_name;
        static std::string rest_postfix;
        static bool top_down_naming; // use topmost directory for playlist naming

    protected:

        void set_father(const SongContainer *new_father) { father = new_father; }
        void set_name(const std::string& name_) { name = name_; }

        void add_son(SongContainerPtr new_son) { sons.push_back(new_son); }
        void add_song(SongPtr new_song) { songs.push_back(new_song); }

        void set_ForceShuffled(bool force_shuffled_) { force_shuffled = force_shuffled_; }
        void set_ForceSorted(bool force_sorted_) { force_sorted = force_sorted_; }

        SongContainerList::const_iterator son_begin() const { return sons.begin(); }
        SongContainerList::const_iterator son_end() const { return sons.end(); }
        SongContainerList::iterator son_begin() { return sons.begin(); }
        SongContainerList::iterator son_end() { return sons.end(); }

        void merge(SongContainerPtr other);

    public:
        SongContainer(const SongContainer *father_) : force_sorted(false), force_shuffled(false), father(father_), name("") {}
        virtual ~SongContainer() {}

        static void init(const std::string& allrest_def, bool top_down_naming_);

        const SongList& get_list() const;
        std::string get_name() const { return name; }
        std::string get_recursive_name() const;

        bool savePlaylists(SongList& all_songs, size_t min_playlist_size) const;

        void warn_broken_links();
    };


    // ----------------------------------------------
    //      class SongDir : public SongContainer
    // ----------------------------------------------
    class SongDir : public SongContainer {
    private:
        std::string dirname;

    protected:
        const std::string& get_dirname() const { return dirname; }

    public:
        SongDir(const std::string& dirname_, const SongContainer *father_);
        virtual ~SongDir() {}

        static bool ignore_links; // true   = > links are ignored ; false = > target of link is used
        static bool read_info;  // true     = > scan mp3s for ID3 tags
        static bool scan_for_id3v2; // true = > scan for ID3 v2 tags
    };


    // -----------------------------------------------
    //      class AllSongs : public SongContainer
    // -----------------------------------------------
    class AllSongs : public SongContainer { // the global song container
    private:
        std::list<std::string> directories;

    public:
        AllSongs() : SongContainer(0) {}
        virtual ~AllSongs() {}

        void add_directory(const std::string& dir);
    };


}; // end of namespace rs

#else
#error SongContainer.h included twice
#endif // SONGCONTAINER_H

