/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// Time-stamp: <Sun Feb/02/2003 09:42 MET Coder@ReallySoft.de>
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
/////////////////////////////////////////////////////////////////////////////

#define __STL_NO_DRAND48

#include <Option.h>
#include <Tools.h>
#include <oem2ansi.h>
#include <Directory.h>

#include <vector>
#include <functional>

#include <cstdio>
#include <unistd.h>

#include "save.h"
#include "SongContainer.h"
#include "DirContent.h"


// --------------------------------------------------------------------------------

using namespace std;
using namespace rs;
using namespace rs::cmdline;
using namespace rs::err;
using namespace rs::str;
using namespace rs::file;
using namespace rs::posix;
using namespace playlist;

namespace rs {
    bool show_debug_info = false;
};

// --------------------------------------------------------------------------------

ProgramInfo playlistInfo("playlist", 2, 12,
                         "Search directory trees for MP3 files and automatically create playlists for "
                         "WINAMP and other MP3 players.");

// --------------------------------------------------------------------------------

CmdLineArgument playlistDir("pldir", "directory where playlists will be created.");

RepeatedCmdLineArgument rootDir("tree",
                                "root-directorie(s) of trees containing MP3s. It is assumed that your MP3s "
                                "are categorized in sub-directories.",
                                acceptAnyArgument);

// --------------------------------------------------------------------------------

CmdLineOption extendedPlaylist('x', "Use extended WinAmp playlist format (scans for ID3-tags; only works for mp3-files).");
CmdLineOption dontScanID3v2('2', "Don't scan for ID3v2-tags (if you encounter problems)");
CmdLineOption ID3v1_preferred('o', "if ID3v1 and ID3v2 present, then use v1 (default: use v2)");

CmdLineOption sortLists('s', "Sort created playlists by MP3 name.");
CmdLineOption shuffleLists('S', "Shuffle created playlists.");

CmdLineAnnotation about_sorting("If no sort/shuffle flag is specified the found file order is used.\n"
                                "If artist or album directory contains a file named '" SORT_FILE_NAME "' the "
                                "corresponding playlist will be sorted regardless of the specified sort "
                                "order. (analog if it contains a file named '" SHUFFLE_FILE_NAME "')");

CmdLineArgumentOption minsize('m', "size", "minimum size for playlists (playlists with less songs are not saved seperately)", "5");
CmdLineOption         showsize('n', "Show size of playlist in playlist name");

CmdLineOption         byTagPlaylists('t', "Create 'by Tag' playlists");
CmdLineArgumentOption minsizeByTag('M', "size", "minimum size for 'by Tag' playlists.", "1");
CmdLineArgumentOption minsizeByTagSongname('N', "size", "minimum size for 'by Tag/Songname' playlists.", "2");

CmdLineArgumentOption userPlaylistFormat('u', "form", "User-defined playlist entry format (form see -X)", "$p");
CmdLineArgumentOption extendedPlaylistFormat('X', "form",
                                             "Specify extended playlist entry ($s=songtitle $a=artist $b=album "
                                             "$y=year $t=track(id3v1:from comment) $f=file- or linkname $p=fullpath $q=quality "
                                             "$(...)$=only inserted if ... contains non-empty macro)",
                                             "$f");

CmdLineOption clearPlaylists('c', (string("Recursively delete all playlists from ")+
                                   playlistDir.get_name()+
                                   " before creating new playslists").c_str());

CmdLineOption ignoreWindowsLinks('l', "Ignore windows links [Default: Examine links and add target to playlist]");
CmdLineOption skipBrokenLinks('B', "Ignore broken links (no warning)");

CmdLineAnnotation about_links("When playlist reports broken links it normally helps to double click them - "
                              "then the Windows-Explorer corrects these broken links.");

CmdLineOption relativePaths('r', "Use relative paths in playlists (i.e. if you want to put playlists on a CD)");

CmdLineArgumentOption mp3_extensions('e', "ext-lst", "specify a list of file-extensions to scan for", "mp3;ogg");
CmdLineArgumentOption allrest_def('a', "name-def", "specify used names for all-, rest-list and rest-postfix", "-All;-Rest; (rest)");
CmdLineOption top_down_naming('T', "Generate top-down playlist names");

CmdLineOption showDebugInfo('d', "print debugging info");

// eventuell doch wieder eingebauen ?
//
// CmdLineOption allowCorrupt('C', "save corrupt MP3-files in playlists (default: skip them)");
// CmdLineArgumentOption writeID3info('i', "dirname", "Write infos about missing ID3-Tags to <dirname>.");


//  -------------------------------------
//      static void check_switches()
//  -------------------------------------
static void check_switches() {
    playlistInfo.exclude(sortLists, shuffleLists);

//     playlistInfo.exclude(ignoreWindowsLinks, guessBrokenLinks);
    playlistInfo.exclude(ignoreWindowsLinks, skipBrokenLinks);

    playlistInfo.neededBy(extendedPlaylist, dontScanID3v2);
    playlistInfo.neededBy(extendedPlaylist, ID3v1_preferred);
    playlistInfo.neededBy(extendedPlaylist, extendedPlaylistFormat);
    playlistInfo.neededBy(extendedPlaylist, byTagPlaylists);

    playlistInfo.neededBy(byTagPlaylists, minsizeByTag);
}

// -----------------------------------------------------------------------------
//      static void deleteExistingPlaylists(const string& playlist_rootdir)
// -----------------------------------------------------------------------------
static void deleteExistingPlaylists(const string& playlist_rootdir) {
    hasExtension pred_playlist(PLAYLIST_EXTENSION);
    strlist      existing_playlists;

    show_dirs_read = false;
    getDirContentRecursive(playlist_rootdir, pred_playlist, existing_playlists);

    if (!existing_playlists.empty()) {
        cout << "Deleting " << existing_playlists.size() << " existing playlists " << flush;
        for (strlist::const_iterator pl = existing_playlists.begin(); pl != existing_playlists.end(); ++pl) {
            int res = unlink(pl->c_str());
            if (res<0) throw Error(strf("Can't delete '%s'", pl->c_str()));
        }
        cout << "done\n" << flush;
    }
}

// -----------------------------------------------------------------
//      inline void setPlaylistDirectory(const string& dirname)
// -----------------------------------------------------------------
inline void setPlaylistDirectory(const string& dirname) {
    string cdir = correctDir(dirname);

    saveSettings.playlistDir        = cdir;
    saveSettings.playlistDirCreated = false;
//     create_directory(cdir);
}

typedef map<string, SongList> MappedList;

// -------------------------------------------------------------------------------------------------------------------------------------
//      static bool saveByTag(const string& playlist_dir, const string& subdir, const MappedList& mapped, size_t min_playlist_size)
// -------------------------------------------------------------------------------------------------------------------------------------
static bool saveByTag(const string& playlist_dir, const string& subdir, const MappedList& mapped, size_t min_playlist_size) {
    cout << "done\n" << "Saving " << subdir << " playlists " << flush;

    setPlaylistDirectory(playlist_dir+subdir);

    SongList various;
    bool     saved_sth = false;

    for (MappedList::const_iterator mli = mapped.begin(); mli != mapped.end(); ++mli) {
        const string&   tag_content = mli->first;
        const SongList& songs       = mli->second;
        bool            saved       = false;

        if (songs.size() >= min_playlist_size) saved = savePlaylist(tag_content, songs, false, false, min_playlist_size);
        if (saved) {
            saved_sth = true;
        }
        else {
            COPY_ALL(songs, various);
        }
    }

    if (!various.empty()) {
        if (savePlaylist("[ Various ]", various, false, false, min_playlist_size)) saved_sth = true;
    }

    return saved_sth;
}

//  -----------------------------------------------
//      int main(int argc, char *argv[])
//  -----------------------------------------------
int main(int argc, char *argv[]) {
    try {
        bool have_min_size = false;

        playlistInfo.init(argc, argv);
        check_switches();

        initPredicates(mp3_extensions.get());
        try {
            SongContainer::init(allrest_def.get(), top_down_naming);
        }
        catch (Error& err) {
            throw Error(strf("%s in argument to -%c", allrest_def.get_optChar()));
        }

        if (showDebugInfo) show_debug_info = true;

        string playlist_rootdir  = canonicalize(correctDir(playlistDir.get()));
        size_t min_playlist_size = minsize.getNumeric();

        if (min_playlist_size>1) have_min_size = true;

        MPEG_Info::prefer_id3v2 = !ID3v1_preferred;

        srand(time(0));

        SongDir::ignore_links   = ignoreWindowsLinks;
        SongDir::read_info      = extendedPlaylist;
        SongDir::scan_for_id3v2 = !dontScanID3v2;

        // Scan directories for MP3s :

        cout << "Reading directories " << flush;
        AllSongs all;
        while (rootDir) {
            string dirname = canonicalize(rootDir.get());
            if (show_debug_info) cout << "Reading root directory '" << dirname << "'\n";
            all.add_directory(dirname);
        }
        cout << "done\n" << flush;

        // Delete existing playlists

        create_directory(playlist_rootdir);
        if (clearPlaylists) deleteExistingPlaylists(playlist_rootdir);

        if (!skipBrokenLinks) {
            all.warn_broken_links();
        }

        // Save Playlists :

        saveSettings.ignoreWindowsLinks     = ignoreWindowsLinks;
        saveSettings.showsize               = showsize;
        saveSettings.generally_sort         = sortLists;
        saveSettings.generally_shuffle      = shuffleLists;
        saveSettings.extended_winamp_format = extendedPlaylist;
        saveSettings.normal_entry_pattern   = userPlaylistFormat.get();
        saveSettings.extended_entry_pattern = extendedPlaylistFormat.get();
        saveSettings.relativePaths          = relativePaths;

        cout << "Saving playlists " << flush;

        setPlaylistDirectory(playlist_rootdir);

        SongList allSongs;
        bool     saved_sth = all.savePlaylists(allSongs, min_playlist_size);

        if (extendedPlaylist && byTagPlaylists) {
            MappedList by_artist;
            MappedList by_year;
            MappedList by_songname;
            MappedList by_genre;
            MappedList by_comment;

            MappedList by_bitrate;
            MappedList by_duration;
            MappedList by_warnings;

            MappedList by_orig_artist;
            MappedList by_copyright;
            MappedList by_encoder;
            MappedList by_url;

            MappedList by_conductor;
            MappedList by_composer;
            MappedList by_orchestra;
            MappedList by_opus;

            for (SongList::iterator si = allSongs.begin(); si != allSongs.end(); ++si) {
                SongPtr     sp    = *si;
                const Song& song = *sp;

                if (!song.isLink() && song.has_MPEG_info()) {
                    const MPEG_Info& info = song.get_MPEG_Info();

                    size_t wcount = info.no_of_warnings();
                    if (wcount) { // mpeg info contains warnings
                        for (size_t idx = 0; idx<wcount; ++idx) {
                            by_warnings[info.get_warning(idx)].push_back(sp);
                        }
                    }

                    if (info.infoError()) {
                        by_warnings[string("info-error: ")+info.getErrorReason()].push_back(sp);
                        continue;
                    }

                    const string&   artist   = info.get_artist();
                    const string&   year     = info.get_year();
                    const string&   songname = info.get_songname();
                    const string&   comment  = info.get_comment();
                    const genreSet& genres   = info.get_genres();

                    const string& orig_artist = info.get_orig_artist();
                    const string& copyright   = info.get_copyright();
                    const string& encoder     = info.get_encoder();
                    const string& url         = info.get_url();

                    const string& conductor = info.get_conductor();
                    const string& composer  = info.get_composer();
                    const string& orchestra = info.get_orchestra();
                    const string& opus      = info.get_opus();

                    by_artist[artist.empty() ? "[ Unknown artist ]" : artist].push_back(sp);
                    by_year[year.empty() ? "[ Unknown year ]" : year].push_back(sp);
                    by_songname[songname.empty() ? "[ Unknown songname ]" : songname].push_back(sp);

                    if (!comment.empty()) by_comment[comment].push_back(sp);
                    if (!url.empty()) by_url[url].push_back(sp);

                    if (!orig_artist.empty()) by_orig_artist[orig_artist].push_back(sp);
                    if (!copyright.empty()) by_copyright[copyright].push_back(sp);
                    if (!encoder.empty()) by_encoder[encoder].push_back(sp);

                    if (!conductor.empty()) by_conductor[conductor].push_back(sp);
                    if (!composer.empty()) by_composer[composer].push_back(sp);
                    if (!orchestra.empty()) by_orchestra[orchestra].push_back(sp);
                    if (!opus.empty()) by_opus[opus].push_back(sp);

                    if (info.is_VBR()) {
                        int approxBitrate  = ((info.get_bitrate()+5)/10)*10;
                        by_bitrate[strf("VBR approx. %i kbps", approxBitrate)].push_back(sp);
                    }
                    else {
                        by_bitrate[strf("%i kbps", info.get_bitrate())].push_back(sp);
                    }

                    {
                        int approxDuration = ((info.get_duration()+15)/30)*30;
                        int minutes        = approxDuration/60;
                        int seconds        = approxDuration-60*minutes;

                        by_duration[strf("approx. %3im %02is", minutes, seconds)].push_back(sp);
                    }

                    bool added_to_genre = false;
                    for (genreSet::const_iterator gsi = genres.begin(); gsi != genres.end(); ++gsi) {
                        const string& genre = *gsi;
                        if (genre.length()) {
                            by_genre[genre].push_back(sp);
                            added_to_genre = true;
                        }
                    }
                    if (!added_to_genre) by_genre["[ Unknown genre ]"].push_back(sp);

                    by_warnings[strf("[info stereo] %s", info.get_stereo_mode_name())].push_back(sp);
                }
            }

            size_t min_tag_playlist_size          = minsizeByTag.getNumeric();
            size_t min_tag_songname_playlist_size = minsizeByTagSongname.getNumeric();
            string tagRootDir                     = correctDir(playlist_rootdir + "by Tag");

            if (min_tag_playlist_size>1 || min_tag_songname_playlist_size>1) have_min_size = true;

            create_directory(tagRootDir);

            if (saveByTag(tagRootDir, "[Warning]", by_warnings, min_tag_playlist_size)) saved_sth  = true;
            if (saveByTag(tagRootDir, "[Bitrate]", by_bitrate, min_tag_playlist_size)) saved_sth   = true;
            if (saveByTag(tagRootDir, "[Duration]", by_duration, min_tag_playlist_size)) saved_sth = true;

            if (saveByTag(tagRootDir, "Year", by_year, min_tag_playlist_size)) saved_sth                  = true;
            if (saveByTag(tagRootDir, "Songname", by_songname, min_tag_songname_playlist_size)) saved_sth = true;
            if (saveByTag(tagRootDir, "Genre", by_genre, min_tag_playlist_size)) saved_sth                = true;
            if (saveByTag(tagRootDir, "Comment", by_comment, min_tag_playlist_size)) saved_sth            = true;

            if (saveByTag(tagRootDir, "Artist (original)", by_orig_artist, min_tag_playlist_size)) saved_sth = true;
            if (saveByTag(tagRootDir, "Copyright", by_copyright, min_tag_playlist_size)) saved_sth           = true;
            if (saveByTag(tagRootDir, "Encoder", by_encoder, min_tag_playlist_size)) saved_sth               = true;
            if (saveByTag(tagRootDir, "URL", by_url, min_tag_playlist_size)) saved_sth                       = true;

            if (saveByTag(tagRootDir, "Classical Conductor", by_conductor, min_tag_playlist_size)) saved_sth = true;
            if (saveByTag(tagRootDir, "Classical Composer", by_composer, min_tag_playlist_size)) saved_sth   = true;
            if (saveByTag(tagRootDir, "Classical Orchestra", by_orchestra, min_tag_playlist_size)) saved_sth = true;
            if (saveByTag(tagRootDir, "Classical Opus", by_opus, min_tag_playlist_size)) saved_sth           = true;

            if (saveByTag(tagRootDir, "Artist", by_artist, min_tag_playlist_size)) saved_sth = true;
        }

        cout << "done\n" << flush;
        if (!saved_sth) {
            if (have_min_size) {
                throw Error("No playlists were saved. There were no or not enough input files. "
                            "Check all switches handling 'minimum size' of playlists.");
            }
            else {
                throw Error("No playlists were saved, because no input files were found.");
            }
        }

        return EXIT_SUCCESS;
    }
    catch (Error &err) {
        err.print(cout);
    }
    catch (...) {
        Error err("Unknown exception");
        err.print(cout);
    }

    return EXIT_FAILURE;
}

